--[[
		    ----o----(||)----o----(||)----o----(||)----o----(||)----o----(||)----o----(||)----o----(||)----o----

                        #######      ####    #######    ########         #           ####    #           #  ########
                        #     ##    ##  ##   #     ##   #                #         ##    ##  ##         ##  #
                        #       #  #      #  #       #  #                #        ##      ##  ##       ##   #
                        #     ##   #      #  #     ##   #                #        #        #   #       #    #
                        #######    ########  #######    ########         #        #        #   ##     ##    ########
                        #   #      #      #  #   #      #                #        #        #    ##   ##     #
                        #    #     #      #  #    #     #                #        ##      ##     #   #      #
                        #     #    #      #  #     #    #                #         ##    ##      ## ##      #
                        #      #   #      #  #      #   ########         ########    ####          #        ########

		    ----o----(||)----o----(||)----o----(||)----o----(||)----o----(||)----o----(||)----o----(||)----o----


						    ----o----(||)----oo----(||)----o----

							  v1.10 3rd September 2017
						    Copyright (C) Taraezor / Chris Birch

						    ----o----(||)----oo----(||)----o----

]]
--------------------------------------------------------------------------------------------------------------------------------------------

local addonName, L = ...
local macroName = L["RareLove"]
local rareLoveButtons = {}

local killIndex, cataIndex, extrasIndex, faction, class= 0, 0, 0, "", ""

local select = select
local find = string.find
local gmatch = string.gmatch
local gsub = string.gsub
local lower = string.lower
local strsplit = strsplit
local sub = string.sub
local tostring = tostring
local tremove = tremove

local CreateMacro = CreateMacro
local DoEmote = DoEmote
local GetAchievementCriteriaInfo = GetAchievementCriteriaInfo
local GetAchievementInfo = GetAchievementInfo
local GetAchievementLink = GetAchievementLink
local GetAchievementNumCriteria = GetAchievementNumCriteria
local GetAddOnMetadata = GetAddOnMetadata
local GetFrameRate = GetFrameRate
local GetMacroIndexByName = GetMacroIndexByName
local GetNumMacros = GetNumMacros
local GetTime = GetTime
local InCombatLockdown = InCombatLockdown
local IsSpellKnown = IsSpellKnown
local UnitLevel = UnitLevel
local PlaySound = PlaySound
local StopSound = StopSound
local UnitClass = UnitClass
local UnitExists = UnitExists
local UnitGUID = UnitGUID
local UnitIsBattlePet = UnitIsBattlePet
local UnitIsDeadOrGhost = UnitIsDeadOrGhost
local UnitName = UnitName

--------------------------------------------------------------------------------------------------------------------------------------------
-- Print Colours - 
--------------------------------------------------------------------------------------------------------------------------------------------

local pc_colour_Prefix		= "\124cFFCD853F"
local pc_colour_Highlight	= "\124cFFDEB887"
local pc_colour_PlainText	= "\124cFFCD5C5C"


local function printPC( message )
	DEFAULT_CHAT_FRAME:AddMessage( pc_colour_Prefix.. addonName.. ": ".. pc_colour_PlainText.. message.. "\124r" )
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Tables
--------------------------------------------------------------------------------------------------------------------------------------------

-- The MoP mount and pet drop rares are included. Unable to reliably test for ownership due to API limitations

extraNPCsTable = {	-- MoP Non-achievement (rare) elites
			L["Gokk'lok"],			-- Toy
			L["Huggalon the Heart Watcher"],-- Toy
			L["Martar the Not-So-Smart"],
			L["Fixxul Lonelyheart"],
			L["Lorbu Sadsummon"],
			L["Feverbite"],
			L["Zhing"],
			L["Pengsong"],
			L["Sungraze Behemoth"],
			L["Wild Onyx Serpent"],
			L["Scotty"],
			L["Captain Zvezdan"],		-- Required for Rattleskew in the Timeless Champion achievement
			L["Degu"],			-- Hunter Pet
			L["Gumi"],			-- Hunter Pet
			L["Hutia"],			-- Hunter Pet
			L["Nalak"],			-- Mount
			L["Oondasta"],			-- Mount
			L["Sha of Anger"],		-- Mount
			L["G'nathus"],			-- Pet
			L["Alani"],			-- Mount

			L["Lil Timmy"],			-- Stormwind. Sells a White Kitten companion
			L["The Mad Merchant"],		-- The Wonderworks Toy Shop in Legion Dalaran. Toy / bag / companion / mount
							-- Could not find a few translations
			L["The Amazing Zanzo"],		-- Required by mages to learn Arcane Momentum

			L["Darkmoon Rabbit"],		-- Darkmoon Faire
			L["Moonfang"],

			L["Treasure Goblin"],		-- Special Diablo World Event. Not translated by Blizzard
			L["The Cow King"],
			L["Omen"],			-- Lunar Festival

			L["Bro'Gaz the Clanless"],	-- Nagrand

			L["Doctor Theolen Krastinov"],	-- Scholomance Heroic rare. Toy drop

			L["Time-Lost Proto-Drake"],	-- The Storm Peaks
			L["Skoll"],
			L["Gondria"],			-- Zul'Drak
			L["Arcturis"],			-- Grizzly Hills
}

cataNPCsTable = {	-- Blizzard didn't give these their own achievement, thus they need to be individually listed
			L["Burgy Blackheart"],		-- Vashj'ir
			L["Poseidus"],
			L["Ghostcrawler"],
			L["Mobus"],
			L["Shok'sharak"],
			L["Lady La-La"],
			L["Aeonaxx"],			-- Deepholm
			L["Golgarok"],
			L["Jadefang"],
			L["Terborus"],
			L["Xariona"],
			L["Ankha"],			-- Mount Hyjal
			L["Ban'thalos"],
			L["Blazewing"],
			L["Garr"],
			L["Magria"],
			L["Terropene"],
			L["Thartuk the Exile"],
			L["Akma'hat"],			-- Uldum
			L["Armagedillo"],
			L["Cyrus the Black"],
			L["Madexx"],
			L["Julak-Doom"],		-- Twilight Highlands
			L["Karoma"],
			L["Overlord Sunderfury"],
			L["Sambas"],
			L["Tarvus the Vile"],
}

loveAchievementTable = {
			1206,	-- To All The Squirrels I've Loved Before
			2557,	-- To All The Squirrels Who Shared My Life
			5548,	-- To All the Squirrels Who Cared for Me
			6350,	-- To All the Squirrels I Once Caressed?
}

killAchievementTable = {	-- NOTE: FIXED ORDER

			2556,	-- Pest Control. Keep at position #1 here
			1312,	-- Bloody Rare
			2257,	-- Frostbitten
			4975,	-- From Hell's Heart I Stab at Thee (i.e. Whale Shark), Patch 4.0.3
			7439,	-- Glorious -- ~30 x 56 = 1680 is > than the 1023 macroText limit so split into two, Patch 5.0.4
			7439,
			7932,	-- I'm In Your Base, Killing Your Dudes, Patch 5.1
			8078,	-- Zul'Again, Patch 5.2.0
			8103,	-- Champions of Lei Shen (Isle of Thunder), Patch 5.2.0
			8101,	-- It Was Worth Every Ritual Stone (Isle of Thunder), Patch 5.2.0
			8714,	-- Timeless Champion (Timeless Isle), Patch 5.4.0
}

--------------------------------------------------------------------------------------------------------------------------------------------
-- Setup Buttons
--------------------------------------------------------------------------------------------------------------------------------------------

RareLoveFoundOne = false

local function FillLoveButtons( t, y )

	local name, completed, b, c, macroText = "", false, 0, 0, "/stopmacro [exists]\n"

	local numCriteria = GetAchievementNumCriteria( y )

	for i = 1, numCriteria do
		name, _, completed = GetAchievementCriteriaInfo( y, i )
		if ( completed == false ) then
			macroText = macroText.. "/targetexact ".. name.. "\n"
			b = b + 1
		else
			c = c + 1
		end
	end

	if ( b > 0 ) then
		macroText = macroText.. "/run RareLoveTargetLove(".. t.. ",".. y.. ")\n"
		rareLoveButtons[ t ]:SetAttribute( "macrotext", macroText )
	else
		rareLoveButtons[ t ]:SetAttribute( "macrotext", "" )
	end
end

local function FillKillButtons( t, y )

	local name, completed, b, macroText, level = "", false, 0, "/stopmacro [exists]\n", UnitLevel( "player" )

	if  ( ( t > 2 ) and ( level < 65 ) ) or ( ( t > 3 ) and ( level < 75 ) ) or ( ( t > 4 ) and ( level < 80 ) ) then
		rareLoveButtons[ killIndex+t ]:SetAttribute( "macrotext", "" )
		return
	end

	if ( y == 4975 ) then
		local whaleDone = select( 13, GetAchievementInfo( 4975 ) )
		if ( whaleDone == false ) then
			macroText = macroText.. "/targetexact ".. L["Whale Shark"].. "\n"
			macroText = macroText.. "/run RareLoveTargetKill(".. t.. ",".. y.. ")\n"
			rareLoveButtons[ killIndex+t ]:SetAttribute( "macrotext", macroText )
		else
			rareLoveButtons[ killIndex+t ]:SetAttribute( "macrotext", "" )
		end
		return
	end

	local numCriteria = GetAchievementNumCriteria( y )
	local start = ( t == 6 ) and 29 or 1

	for i = start, numCriteria do
		name, _, completed, _, _, _, _, assetID = GetAchievementCriteriaInfo( y, i )
		if ( y == 8078 ) then
			if ( i == 1 ) then
				-- Patrol paths. Mount drop - unable to reliably test for completion
				macroText = macroText.. "/targetexact ".. L["Zandalari Warscout"].. "\n"
			else
				-- Spawns in very specific locations. Mount drop - unable to reliably test for completion
				macroText = macroText.. "/targetexact ".. L["Zandalari Warbringer"].. "\n"
			end
			b = b + 1
		elseif ( y == 7932 ) then
			-- Spawn in very specific locations. You will only be here if looking for it
			-- Not sure what players prefer here. You woudln't be killing these for the drops, lol
			if ( i == 1 ) then
				if ( faction == "Alliance" ) then
					macroText = macroText.. "/targetexact ".. L["Kar Warmaker"].. "\n"
				else
					macroText = macroText.. "/targetexact ".. L["Dalan Nightbreaker"].. "\n"
				end
			elseif ( i == 2 ) then
				if ( faction == "Alliance" ) then
					macroText = macroText.. "/targetexact ".. L["Ubunti the Shade"].. "\n"
				else
					macroText = macroText.. "/targetexact ".. L["Mavis Harms"].. "\n"
				end
			else
				if ( faction == "Alliance" ) then
					macroText = macroText.. "/targetexact ".. L["Muerta"].. "\n"
				else
					macroText = macroText.. "/targetexact ".. L["Disha Fearwarden"].. "\n"
				end
			end
			b = b + 1
		elseif ( y == 2556 ) then
			-- Pest Control. Problem of needing to show player updated progress. Very unlikely other dead critters in
			-- the area so targeting all unlikely to first locate a dead one
			if ( completed == false ) then
				macroText = macroText.. "/targetexact ".. name.. "\n"
				b = b + 1
			end
		elseif ( y == 7439 ) then
			-- Some drop Toys I can test for. Others fun / useful effects. Chance of bags with AH / vendor mats
			-- Gear/invites are BoP but vendor gold is okay. Decided to ignore completion as the drops are worthwhile

			if ( t == 5 ) and ( i > 28 ) then break end
			macroText = macroText.. "/targetexact ".. name.. "\n"
			b = b + 1
		else
			-- Frostbitten / Bloody Rare. Blue drops are BoE. Nice amount of gold/cloth. Decided to ignore completion
			-- Whale Shark needed dead to help with Respawn timing
			macroText = macroText.. "/targetexact ".. name.. "\n"
			b = b + 1
		end
	end

	if ( b > 0 ) then
		macroText = macroText.. "/run RareLoveTargetKill(".. t.. ",".. y.. ")\n"
		rareLoveButtons[ killIndex+t ]:SetAttribute( "macrotext", macroText )
	else
		rareLoveButtons[ killIndex+t ]:SetAttribute( "macrotext", "" )
	end
end

local function FillCataButton()

	local macroText = "/stopmacro [exists]\n"

	for i, v in ipairs( cataNPCsTable ) do
		macroText = macroText.. "/targetexact ".. v.. "\n"
	end
	macroText = macroText.. "/run RareLoveTargetCata()\n"
	rareLoveButtons[ cataIndex ]:SetAttribute( "macrotext", macroText )
end

local function FillExtrasButton()

	local macroText = "/stopmacro [exists]\n"

	for i, v in ipairs( extraNPCsTable ) do
		if ( v == L["The Amazing Zanzo"] ) then
			if ( class == "MAGE" ) and not IsSpellKnown( 56384 ) then	-- Arcane Momentum. New for Patch 7.1
				macroText = macroText.. "/targetexact ".. v.. "\n"
			end
		else
			macroText = macroText.. "/targetexact ".. v.. "\n"
		end
	end
	macroText = macroText.. "/run RareLoveTargetExtras()\n"
	rareLoveButtons[ extrasIndex ]:SetAttribute( "macrotext", macroText )
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Setup & Progress
--------------------------------------------------------------------------------------------------------------------------------------------

local canFillScripts, completelyFilled, updateRequested, stepFill  = false, false, false, 0

local function FillScripts()

	local totalLove, totalKill, totalCata = 0, 0, 0

	for i, v in ipairs( loveAchievementTable ) do
		if ( i > stepFill ) then break end
		if ( i == stepFill ) then
			FillLoveButtons( i, v )
		end
		totalLove = i
	end
	if ( stepFill > totalLove ) then
		for i, v in ipairs( killAchievementTable ) do
			if ( (i+totalLove) > stepFill ) then break end
			if ( i == (stepFill-totalLove) ) then
				FillKillButtons( i, v )
			end
			totalKill = i
		end
	end
	if ( stepFill > (totalLove+totalKill) ) then
		if ( stepFill == (totalLove+totalKill+1) ) then
			FillCataButton()
		end
		totalCata = totalLove+totalKill+1
	end
	if ( stepFill > (totalLove+totalKill+totalCata) ) then
		if ( stepFill == (totalLove+totalKill+totalCata+1) ) then
			FillExtrasButton()
		end
	end
	if ( stepFill <= (totalLove+totalKill+totalCata+1) ) then 
		stepFill = stepFill + 1
	else
		completelyFilled = true
	end
end

local function ShowProgress()

	for _, v in ipairs( loveAchievementTable ) do
		local b, numCriteria = 0, GetAchievementNumCriteria( v )
		for i = 1, numCriteria do
			local _, _, completed = GetAchievementCriteriaInfo( v, i )
			if ( completed == false ) then
				b = b + 1
			end
		end
--		printPC( select( 2, GetAchievementInfo( v ) ).. " ".. pc_colour_Highlight.. ( numCriteria-b ).. "/".. numCriteria )
		printPC( GetAchievementLink( v ).. " ".. pc_colour_Highlight.. ( numCriteria-b ).. "/".. numCriteria )
	end

	local level = UnitLevel( "player" )
	for i, v in ipairs( killAchievementTable ) do

		if  ( ( i > 2 ) and ( level < 65 ) ) or ( ( i > 3 ) and ( level < 75 ) ) or ( ( i > 4 ) and ( level < 80 ) ) then

		elseif ( i ~= 6 ) then
			local b, numCriteria = 0, GetAchievementNumCriteria( v )
			for j = 1, numCriteria do
				local _, _, completed = GetAchievementCriteriaInfo( v, j )
				if ( completed == false ) then
					b = b + 1
				end
			end
--			printPC( select( 2, GetAchievementInfo( v ) ).. " ".. pc_colour_Highlight.. ( numCriteria-b ).. "/".. numCriteria )
			printPC( GetAchievementLink( v ).. " ".. pc_colour_Highlight.. ( numCriteria-b ).. "/".. numCriteria )
		end
	end
end

local function ShowProgressPartial( y )

	local b, numCriteria, remaining, criteriaString, completed, reqQuantity = 0, GetAchievementNumCriteria( y ), "", "", false, 0

	for i = 1, numCriteria do
		criteriaString, _, completed, _, reqQuantity = GetAchievementCriteriaInfo( y, i )
		if ( completed == false ) then
			b = b + 1
			remaining = ( remaining == "" ) and criteriaString or ( remaining.. ", ".. criteriaString )
			if ( reqQuantity > 1 ) then
				remaining = remaining.. "(".. pc_colour_Highlight.. reqQuantity.. pc_colour_PlainText.. ")"
			end
		end
	end
	printPC( GetAchievementLink( y ).. " ".. pc_colour_Highlight.. ( numCriteria-b ).. "/".. numCriteria )
	if ( remaining ~= "" ) then
		print( pc_colour_PlainText.. "( ".. remaining.. " )" )
	end
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Global Functions
--------------------------------------------------------------------------------------------------------------------------------------------

local partialLoveUpdate, partialLoveTimeStart, partialKillUpdate, partialKillTimeStart = 0, 0, 0, 0

function RareLoveTargetLove( t, y )

	local name = UnitName( "target" )
	if not name then return end
	local guidType = strsplit( "-", UnitGUID( "target" ) )
	
	local loveName = "/targetexact ".. name.. "\n"
	local macroText = rareLoveButtons[ t ]:GetAttribute( "macrotext" )

	if find( macroText, loveName ) then
		local isBattlePet = UnitIsBattlePet( "target" ) or false
		if ( guidType == "Creature" ) or ( ( guidType == "Pet" ) and ( isBattlePet == true ) ) then
			DoEmote( "love" )
			PlaySoundFile( RareLoveDB.loveSound )
			macroText = gsub( macroText, loveName, "", 1 )
			if not find( macroText, "/targetexact" ) then macroText = "" end
			rareLoveButtons[ t ]:SetAttribute( "macrotext", macroText )
			partialLoveUpdate, partialLoveTimeStart = y, GetTime()
		end
	end
end

local function KillNotification( y )

	if UnitIsDeadOrGhost( "target" ) and ( y ~= 4975 ) then
		partialKillUpdate, partialKillTimeStart = y, GetTime()
	else
		printPC( "\124cFFFF0000".. L["Kill!"] )
		PlaySoundFile( RareLoveDB.killSound )
		if ( y == 2556 ) then
			partialKillUpdate, partialKillTimeStart = y, GetTime() + 20
		end
	end
end

local foundRareFrozen = 0

function RareLoveTargetKill( t, y )

	local name = UnitName( "target" )
	if not name then return end
	local guidType, _, _, _, _, npcID  = strsplit( "-", UnitGUID( "target" ) )
	
	local macroText, killName = rareLoveButtons[ killIndex+t ]:GetAttribute( "macrotext" ), ""
	killName = "/targetexact ".. name.. "\n"

	if find( macroText, killName ) then
		if ( y == 2556 ) then
			if ( guidType == "Pet" ) or ( guidType == "Creature" ) then
				KillNotification( y )
			end
		elseif ( guidType == "Creature" ) then
			if ( y == 4975 ) then
				if ( npcID == 40728 ) then	-- Must test for the exact Whale Shark
					KillNotification( y )
				end
			else
				KillNotification( y )
			end
		end
	end
end

function RareLoveTargetCata()	-- For now consider the Cata as just more Extras

	local name = UnitName( "target" )
	if not name then return end
	local guidType = strsplit( "-", UnitGUID( "target" ) )
	local macroText = rareLoveButtons[ cataIndex ]:GetAttribute( "macrotext" )
	if find( macroText, "/targetexact ".. name.. "\n" ) then
		if ( guidType == "Creature" ) then
			PlaySoundFile( RareLoveDB.extrasSound )
		end
	end
end

function RareLoveTargetExtras()

	local name = UnitName( "target" )
	if not name then return end
	local guidType = strsplit( "-", UnitGUID( "target" ) )
	local macroText = rareLoveButtons[ extrasIndex ]:GetAttribute( "macrotext" )
	if find( macroText, "/targetexact ".. name.. "\n" ) then
		if ( guidType == "Creature" ) then
			PlaySoundFile( RareLoveDB.extrasSound )
		end
	end
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- OnEvent
--------------------------------------------------------------------------------------------------------------------------------------------

local timesIn, login = 0, true

function RareLove_OnEvent( self, event, addonNameEvent )

	if ( event == "UPDATE_MACROS" ) then
		timesIn = timesIn + 1
		if ( timesIn == 2 ) then
			if ( GetMacroIndexByName( macroName ) == 0 ) then
				if ( select( 2, GetNumMacros() ) < MAX_CHARACTER_MACROS ) then
					CreateMacro( "RareLove", "inv_jewelcrafting_crimsonhare", "/click RareLoveButton", true )
				else
					local message = L["Sorry, no space for %p macros"]
					message = gsub( message, "%p", UnitName("player") )
					printPC( message )
				end
			end
		end

	elseif ( event == "ADDON_LOADED" ) and ( addonNameEvent == addonName ) then
		local onOff		= RareLoveDB.on or "y"
		local extrasSound	= RareLoveDB.extrasSound or "Sound\\Events\\gruntling_horn_bb.ogg"
		local loveSound		= RareLoveDB.loveSound or "Sound\\Spells\\Shaysbell.ogg"
		local killSound		= RareLoveDB.killSound or "Sound\\Events\\scourge_horn.ogg"
		local debug		= RareLoveDB.debug or false
		RareLoveDB		= {}
		RareLoveDB.on		= onOff
		RareLoveDB.extrasSound	= extrasSound
		RareLoveDB.loveSound	= loveSound
		RareLoveDB.killSound	= killSound
		RareLoveDB.debug	= debug

		local macroText = ""

		rareLoveButtons.main = CreateFrame( "Button", "RareLoveButton", nil, "SecureActionButtonTemplate" )
		rareLoveButtons.main:SetAttribute( "type", "macro" )

		local i = 0
		for t, y in ipairs( loveAchievementTable ) do
			rareLoveButtons[ t ] = CreateFrame( "Button", "RareLoveButtonLove".. (t), nil, "SecureActionButtonTemplate" )
			rareLoveButtons[ t ]:SetAttribute( "type", "macro" )
			macroText = macroText.. "/click RareLoveButtonLove".. (t).. "\n"
			i = t
		end

		killIndex, i = i, 0

		for t, y in ipairs( killAchievementTable ) do
			rareLoveButtons[ t+killIndex ] = CreateFrame( "Button", "RareLoveButtonKill".. (t+killIndex), nil, "SecureActionButtonTemplate" )
			rareLoveButtons[ t+killIndex ]:SetAttribute( "type", "macro" )
			macroText = macroText.. "/click RareLoveButtonKill".. (t+killIndex).. "\n"
			i = t
		end

		cataIndex = killIndex + i + 1

		rareLoveButtons[ cataIndex ] = CreateFrame( "Button", "RareLoveButtonCata", nil, "SecureActionButtonTemplate" )
		rareLoveButtons[ cataIndex ]:SetAttribute( "type", "macro" )
		macroText = macroText.. "/click RareLoveButtonCata\n"

		extrasIndex = cataIndex + 1

		rareLoveButtons[ extrasIndex ] = CreateFrame( "Button", "RareLoveButtonExtras", nil, "SecureActionButtonTemplate" )
		rareLoveButtons[ extrasIndex ]:SetAttribute( "type", "macro" )
		macroText = macroText.. "/click RareLoveButtonExtras\n"

		rareLoveButtons.main:SetAttribute( "macrotext", macroText )

	elseif ( event == "PLAYER_LOGIN" ) then
		canFillScripts = true
		faction = UnitFactionGroup( "player" )
		class = select( 2, UnitClass( "player" ) )
	end
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- OnEvent
--------------------------------------------------------------------------------------------------------------------------------------------

local diagnosticStart = GetTime()

function RareLove_OnUpdate( self, elapsed )

	self.timeSinceLastUpdate = self.timeSinceLastUpdate + elapsed

	if ( RareLoveDB.on == "n" ) or ( canFillScripts == false ) or InCombatLockdown() then return end

	if ( completelyFilled == false ) then
		if ( updateRequested == true ) then
			while self.timeSinceLastUpdate > ( 1 / GetFramerate() * 5 ) do
				FillScripts()
				self.timeSinceLastUpdate = 0
			end
			if ( completelyFilled == true ) then
				ShowProgress()
				updateRequested = false

				if ( RareLoveDB.debug == true ) and ( completelyFilled == true ) then 
					print("Progress refill took "..(GetTime()-diagnosticStart).."s")
				end

			end
		else
			while self.timeSinceLastUpdate > 1 do
				FillScripts()
				self.timeSinceLastUpdate = 0
			end

			if ( completelyFilled == true ) then
				if ( RareLoveDB.debug == true ) then
					print("Macro (re)fill took "..(GetTime()-diagnosticStart).."s")
				end
				if ( login == true )then
					printPC( L["Ready"] )
					login = false
				end
			end
		end
	else
		if ( self.timeSinceLastUpdate > 60 ) then
			completelyFilled, stepFill, diagnosticStart = false, 0, GetTime()

			if ( RareLoveDB.debug == true ) then
				print("60 second cycle Refill commenced")
			end
		end
	end

	if ( partialLoveUpdate > 0 ) then
		if ( GetTime() - partialLoveTimeStart ) > 1 then
			ShowProgressPartial( partialLoveUpdate )
			partialLoveUpdate = 0
		end
	end
	if ( partialKillUpdate > 0 ) then
		if ( GetTime() - partialKillTimeStart ) > 1 then
			ShowProgressPartial( partialKillUpdate )
			partialKillUpdate = 0
		end
	end
end


--------------------------------------------------------------------------------------------------------------------------------------------
-- Slash Commands
--------------------------------------------------------------------------------------------------------------------------------------------

local soundkitHandle = nil

SLASH_RareLove1, SLASH_RareLove2, SLASH_RareLove3 = "/rarelove", "/rare", "/rl"

SlashCmdList[ "RareLove" ] = function( options )

	local options1

	for v in gmatch( options, "%S+" ) do
		v = lower( v )
		options1 = sub( v,1,1 )
		break
	end

	if ( options1 == "?" ) or ( options1 == nil ) then
		printPC( pc_colour_Highlight.. L["Macro Commands"].. ":\n"..
			pc_colour_Highlight.. "p".. pc_colour_PlainText.. " = ".. L["Show your progress"].. "\n"..
			pc_colour_Highlight.. "o".. pc_colour_PlainText.. " = ".. L["Turn on / off"].. ". (".. L["Toggle"].. ")\n"..
			pc_colour_Highlight.. "e".. pc_colour_PlainText.. " = ".. L["Extras sound"].. ". (".. L["Toggle"].. ")\n"..
			pc_colour_Highlight.. "l".. pc_colour_PlainText.. " = ".. L["/love sound"].. ". (".. L["Toggle"].. ")\n"..
			pc_colour_Highlight.. "k".. pc_colour_PlainText.. " = ".. L["Kill! sound"].. ". (".. L["Toggle"].. ")\n"..
			pc_colour_Highlight.. "v".. pc_colour_PlainText.. " = ".. L["Show the version number"].. "\n")

	elseif ( options1 == "o" ) then
		RareLoveDB.on = ( RareLoveDB.on == "y" ) and "n" or "y"
		if ( RareLoveDB.on == "y" ) then
			printPC( L["On"] )
			FillScripts()
		else
			printPC( L["Off"] )
		end

	elseif ( options1 == "e" ) then
		if RareLoveDB.extrasSound == "Sound\\Events\\gruntling_horn_bb.ogg" then
			RareLoveDB.extrasSound = ""
			printPC( "(".. L["Silent"].. ")" )
		else
			RareLoveDB.extrasSound = "Sound\\Events\\gruntling_horn_bb.ogg"
			printPC( "Sound\\Events\\gruntling_horn_bb.ogg" )
		end
		PlaySoundFile( RareLoveDB.extrasSound )

	elseif ( options1 == "l" ) then
		if RareLoveDB.loveSound == "Sound\\Spells\\Shaysbell.ogg" then
			RareLoveDB.loveSound = "Sound\\Spells\\Fairytale_Sfx_01.ogg"
			printPC( "Sound\\Spells\\Fairytale_Sfx_01.ogg" )
		elseif RareLoveDB.loveSound == "Sound\\Spells\\Fairytale_Sfx_01.ogg" then
			RareLoveDB.loveSound = ""
			printPC( "(".. L["Silent"].. ")" )
		else
			RareLoveDB.loveSound = "Sound\\Spells\\Shaysbell.ogg"
			printPC( "Sound/Spells\\Shaysbell.ogg" )
		end
		PlaySoundFile( RareLoveDB.loveSound )

	elseif ( options1 == "k" ) then
		if RareLoveDB.killSound == "Sound\\Events\\scourge_horn.ogg" then
			RareLoveDB.killSound = "Sound\\Event Sounds\\Event_wardrum_ogre.ogg"
			printPC( "Sound\\Event Sounds\\Event_wardrum_ogre.ogg" )
		elseif RareLoveDB.killSound == "Sound\\Event Sounds\\Event_wardrum_ogre.ogg" then
			RareLoveDB.killSound = ""
			printPC( "(".. L["Silent"].. ")" )
		else
			RareLoveDB.killSound = "Sound\\Events\\scourge_horn.ogg"
			printPC( "Sound\\Events\\scourge_horn.ogg" )
		end
		PlaySoundFile( RareLoveDB.killSound )

	elseif ( options1 == "p" ) then
		completelyFilled, stepFill, updateRequested, diagnosticStart = false, 0, true, GetTime()

	elseif ( options1 == "d" ) then
		RareLoveDB.debug = not RareLoveDB.debug

	elseif ( options1 == "v" ) then
		local version = GetAddOnMetadata( addonName, "Version" )
		printPC( L["Version: "].. pc_colour_Highlight.. version )

	-- Easter Eggs. Enjoy!

	elseif ( options1 == "z" ) then
		if ( soundkitHandle ) then
			StopSound( soundkitHandle, 5000 )
			soundkitHandle = nil
		else
			local soundSelection = random( 1 )
			if ( soundSelection == 1 ) then
				_, soundkitHandle = PlaySound( 78803 )		-- Tristram
			else
				soundSelection = random( 3 )
				if ( soundSelection == 1 ) then
					_, soundkitHandle = PlaySound( 11806 )	-- Tavern Dwarf
				elseif ( soundSelection == 2 ) then
					_, soundkitHandle = PlaySound( 12329 )	-- Tavern Tauren
				else
					_, soundkitHandle = PlaySound( 12137 )	-- Tavern Undead
				end
			end
		end
	end
end